from typing import Literal

from pydantic import BaseModel, Field

MessageRole = Literal["system", "user", "assistant", "function"]


class Message(BaseModel):
    role: MessageRole = Field(
        ...,
        description="The role of the messages author. One of system, user, assistant, or function.",
    )
    content: str = Field(
        ...,
        description="The contents of the message. content is required for all messages, and may be null for assistant messages with function calls.",
    )



from abc import ABC
from typing import Literal

from langchain.llms import HuggingFaceEndpoint

from genoss.entities.chat.chat_completion import ChatCompletion
from genoss.entities.chat.message import Message
from genoss.llm.base_genoss import BaseGenossLLM


class HuggingFaceInferenceEndpointLLM(BaseGenossLLM, ABC):
    """Class for interacting with Hugging Face Inference APIs."""

    # Subclasses must define these
    name = "HF Inference Endpoint"
    api_key: str | None = None
    endpoint_url: str
    description: str = "Hugging Face Inference API custom endpoint."
    task: Literal[
        "text-generation", "text-generation", "summarization"
    ] = "text-generation"

    def generate_answer(self, messages: list[Message]) -> ChatCompletion:
        """Generate answer from prompt."""
        llm = HuggingFaceEndpoint(
            endpoint_url=self.endpoint_url,
            huggingfacehub_api_token=self.api_key,
            task=self.task,
        )
        return self._chat_completion_from_langchain_llm(llm=llm, messages=messages)

    def generate_embedding(self, text: str) -> list[float]:
        """Dummy method to satisfy base class requirement."""
        # TODO: why is this necessary? Architecture issue?
        raise NotImplementedError(
            "This method is not used for Hugging Face Inference API."
        )



from fastapi import Header, HTTPException


class AuthHandler:
    @staticmethod
    async def check_auth_header(
        # TODO: check if this is the correct way to use Header
        authorization: str
        | None = Header(None),  # noqa: B008
    ) -> str | None:
        if authorization is None:
            return None

        components = authorization.split()

        if len(components) != 2 or components[0].lower() != "bearer":
            raise HTTPException(status_code=403, detail="Invalid authorization header")

        api_key = components[1]
        return api_key



from genoss.llm.base_genoss import BaseGenossLLM
from genoss.llm.fake_llm import FAKE_LLM_NAME, FakeLLM
from genoss.llm.hf_hub.base_hf_hub import BaseHuggingFaceHubLLM
from genoss.llm.hf_inference_endpoint.hf_inference_endpoint import (
    HuggingFaceInferenceEndpointLLM,
)
from genoss.llm.local.gpt4all import Gpt4AllLLM
from genoss.llm.openai_llm.openai_llm import OpenAILLM
from genoss.services.model_routing_helpers import ModelRouter

chat_completion_registry = ModelRouter[BaseGenossLLM]()
# Warning: matches is order dependent. The first match will be used.
chat_completion_registry.register_model(
    paths="gpt4all",
    model_class=Gpt4AllLLM,
    call_params_keys={"api_key"},
)
chat_completion_registry.register_model(
    paths=FAKE_LLM_NAME,
    model_class=FakeLLM,
    call_params_keys={"api_key"},
)
chat_completion_registry.register_model(
    paths=["{model_name:str}", "openai/{model_name:str}"],
    model_class=OpenAILLM,
    call_params_keys={"api_key"},
)
chat_completion_registry.register_model(
    # Path catches slashes hf-hub/username/repo_name
    # For example those are valid :
    # hf-hub/gpt-2 and hf-hub/tiiuae/falcon-40b
    paths=["hf-hub/{repo_id:path}"],
    model_class=BaseHuggingFaceHubLLM,
    call_params_keys={"api_key"},
)
chat_completion_registry.register_model(
    paths=["hf-inference-endpoint/{endpoint_url:path}"],
    model_class=HuggingFaceInferenceEndpointLLM,
    call_params_keys={"api_key"},
)



from pathlib import Path

from pydantic import validator

from genoss.llm.base_genoss import BaseGenossLLM


class BaseLocalLLM(BaseGenossLLM):
    model_path: str

    @validator("model_path")
    def validate_model_path(cls, v: str) -> str:
        if not Path(v).is_file():
            raise ValueError(f"The provided model path does not exist: {v}")
        return v



from abc import abstractmethod

from langchain import LLMChain, PromptTemplate
from pydantic import BaseModel

from genoss.entities.chat.chat_completion import ChatCompletion
from genoss.entities.chat.message import Message
from genoss.prompts.prompt_template import prompt_template


class BaseGenossLLM(BaseModel):
    name: str
    description: str

    @abstractmethod
    def generate_answer(self, messages: list[Message]) -> ChatCompletion:
        pass

    def _chat_completion_from_langchain_llm(
        self,
        llm: BaseModel,
        messages: list[Message],
        prompt: PromptTemplate | None = None,
    ) -> ChatCompletion:
        if prompt is None:
            prompt = prompt_template
        llm_chain = LLMChain(prompt=prompt, llm=llm)

        question = messages[-1].content
        response_text = llm_chain(question)

        answer = response_text["text"]

        return ChatCompletion.from_model_question_answer(
            model=self.name, answer=answer, question=question
        )

    @abstractmethod
    def generate_embedding(self, text: str) -> list[float]:
        pass



from fastapi import APIRouter

from genoss.llm.local.gpt4all import Gpt4AllLLM
from logger import get_logger

logger = get_logger(__name__)

embeddings_router = APIRouter()


@embeddings_router.post("/embeddings", tags=["Embeddings"])
async def post_embeddings(
    model: str,
    input: str,
) -> list[float]:
    if model == "gpt4all":
        gpt = Gpt4AllLLM(name="gpt4all")
    else:
        raise NotImplementedError("Model can not be anything else than gpt4all.")

    response = gpt.generate_embedding(input)
    return response



from langchain.chat_models import ChatOpenAI
from langchain.embeddings import OpenAIEmbeddings
from langchain.schema import BaseMessage, ChatMessage
from pydantic import Field

from genoss.entities.chat.chat_completion import ChatCompletion
from genoss.entities.chat.message import Message
from genoss.llm.base_genoss import BaseGenossLLM


class OpenAILLM(BaseGenossLLM):
    name: str = "openai"
    description: str = "OpenAI LLM"
    model_name: str = Field("gpt-3.5-turbo", description="OpenAI model name")
    api_key: str

    @staticmethod
    def _parse_messages_as_chat_messages(messages: list[Message]) -> list[BaseMessage]:
        return [
            ChatMessage(content=message.content, role=message.role)
            for message in messages
        ]

    def generate_answer(self, messages: list[Message]) -> ChatCompletion:
        llm = ChatOpenAI(model_name=self.model_name, openai_api_key=self.api_key)

        chat_messages = self._parse_messages_as_chat_messages(messages)
        response = llm(chat_messages)

        question = messages[-1].content
        answer = response.content

        return ChatCompletion.from_model_question_answer(
            model=self.name, answer=answer, question=question
        )

    def generate_embedding(self, text: str) -> list[float]:
        model = OpenAIEmbeddings(model=self.model_name, openai_api_key=self.api_key)
        return model.embed_query(text)



"""This is highly inspired from starlette.routing.

The goal is to have a ModelRouter that can be used to register models and their routes.
We stick to the starlette.routing API as much as possible to have a known interface.

For the moment, we do not handle nested routes.
"""
import re
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, Field, ValidationError, root_validator
from starlette.convertors import Convertor
from starlette.routing import Match, compile_path

from genoss.llm.base_genoss import BaseGenossLLM


class RouterException(Exception):
    pass


class InvalidRouteParameters(ValueError, RouterException):
    pass


class RouteNotFound(ValueError, RouterException):
    pass


ModelT = TypeVar("ModelT", bound=BaseGenossLLM)
RouteParams = dict[str, Any]
CallParams = dict[str, Any]


class AbstractRoute(ABC):
    """The abstract route.

    Similar to : from starlette.routing import BaseRoute.
    """

    @abstractmethod
    def matches(self, path: str) -> tuple[Match, RouteParams]:
        """Check if the provided path should be handled by this route.

        Return the type of Match and the params found in the route.
        """
        raise NotImplementedError()

    @abstractmethod
    def handle(self, call_params: CallParams, route_params: RouteParams) -> Any:
        raise NotImplementedError()


class StandardModelRoute(AbstractRoute, Generic[ModelT], BaseModel):
    """A route for a specific model.

    Similar to : from starlette.routing import Route
    """

    path: str
    model_class: type[ModelT]
    additional_params: RouteParams = Field(default_factory=dict)
    regex: re.Pattern[str]
    formatted_route: str
    converters: dict[str, Convertor[Any]]
    call_params_keys: set[str]

    class Config:
        # Necessary to allow converters with pydantic
        arbitrary_types_allowed = True

    @classmethod
    def from_route_and_model(
        cls,
        path: str,
        model_class: type[ModelT],
        additional_params: RouteParams | None = None,
        call_params_keys: set[str] | None = None,
    ) -> "StandardModelRoute[ModelT]":
        """Create a route for a specific model.

        This is the reference builder.
        Handle creating the different params from the path.
        """
        if additional_params is None:
            additional_params = {}
        if call_params_keys is None:
            call_params_keys = set()
        regex, formatted_route, converters = compile_path(path)
        return cls(
            path=path,
            model_class=model_class,
            additional_params=additional_params,
            regex=regex,
            formatted_route=formatted_route,
            converters=converters,
            call_params_keys=call_params_keys,
        )

    @root_validator(skip_on_failure=True)
    def validate_route(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate that the route parameters are valid for the model."""
        # TODO: check that again to had route_params, call_params and additional_params
        available_params_at_handle = set(values["converters"].keys()) | set(
            values["additional_params"].keys()
        )
        call_params_keys = values["call_params_keys"]
        model_class = values["model_class"]
        model_fields = model_class.__fields__
        required_fields = {
            name for name, field in model_fields.items() if field.required
        }

        non_existing_fields = available_params_at_handle - model_fields.keys()
        if non_existing_fields:
            raise ValueError(
                f"Route variables {non_existing_fields} do not exist in model: {model_class}."
            )
        if required_fields - available_params_at_handle - call_params_keys:
            raise ValueError(
                f"Required fields {required_fields} are not "
                f"in route variables: {available_params_at_handle} "
                f"or call_params_keys: {call_params_keys}."
            )
        return values

    def matches(self, path: str) -> tuple[Match, RouteParams]:
        """Check if the provided path should be handled by this route.

        This is based on the regex created from the path.
        For example, openai/{model_name:str} should :
        - match openai/gpt-3.5-turbo
          with a return value of (Match.FULL, {"model_name": "gpt-3.5-turbo"})
        - not match hf-local/gpt-2
          with a return value of (Match.NONE, {})
        """
        # See later to implement partial routing
        # If we have Router of Router for example
        match = self.regex.match(path)
        if match is None:
            return Match.NONE, {}
        matched_params = match.groupdict()
        converted_params = {
            key: self.converters[key].convert(value)
            for key, value in matched_params.items()
        }
        return Match.FULL, converted_params

    def handle(self, call_params: CallParams, route_params: RouteParams) -> ModelT:
        """Create the Model (__init__) from all the params.

        There are 4 types of parameters that will enter the `model_class.__init__`.
        - route_params (from the route).
          Example: "openai/{model_name:str}-{version:str}" with "openai/gpt-4-0314"
          should result in {"model_name": "gpt-4", "version": "0314"}.
        - additional_params which comes from registering the route.
          Example:```python
              embeddings.register_model(
                "openai/gpt-4",
                OpenAILLM,
                additional_params={"model_name": "gpt-4-0314"},
                call_params={"api_key"}
              )
              ```
        - default values from the model init which does not appear here
          as they are handled by the model class itself.
        - handle_time_params which are params that are not in the route
          but are added at handle time.
        """
        try:
            if call_params.keys() != self.call_params_keys:
                # TODO: create a stronger signature / type system ?
                raise TypeError(
                    f"call_params should contain the same keys than call_params_keys "
                    f"{call_params} != {self.call_params_keys}."
                )
            return self.model_class(
                **call_params, **route_params, **self.additional_params
            )
        except ValidationError as exception:
            # TODO: We should probably think about a better exception
            # that will raise a validation issue to genoss client.
            # Examples could be, the model is not available
            # for your current API key,
            # or the model is not available anymore (deprecation).
            raise InvalidRouteParameters(
                f"Failed to instantiate the model {self.model_class.__name__} "
                f"from the provided parameters {call_params} {route_params} and {self.additional_params}."
            ) from exception


class ModelRouter(Generic[ModelT], BaseModel):
    """The model router.

    Similar to : from starlette.routing import Router

    This is a model registry.
    You can call register to register a model.
    For example :
    ```python
    chat_completion = ModelRouter()
    chat_completion.register(["{model_name:str}", "openai/{model_name:str}"], OpenAILLM)
    chat_completion.register(
       "openai",
       OpenAILLM,
       additional_params={"model_name": "gpt-3.5-turbo"}
    )

    instantiated_model = chat_completion.build_model("openai", {"api_key": "1234"})
    instantiated_model.generate_answer("What is the meaning of life?")
    ```
    """

    routes: list[StandardModelRoute[ModelT]] = Field(default_factory=list)

    def register_model(
        self,
        paths: str | Sequence[str],
        model_class: type[ModelT],
        additional_params: RouteParams | None = None,
        call_params_keys: set[str] | None = None,
    ) -> None:
        if additional_params is None:
            additional_params = {}
        if call_params_keys is None:
            call_params_keys = set()
        if isinstance(paths, str):
            paths = [paths]

        for path in paths:
            route = StandardModelRoute.from_route_and_model(
                path=path,
                model_class=model_class,
                additional_params=additional_params,
                call_params_keys=call_params_keys,
            )
            self.routes.append(route)

    def build_model(self, path: str, call_params: CallParams) -> ModelT:
        for route in self.routes:
            match, route_params = route.matches(path)
            if match is Match.FULL:
                return route.handle(call_params, route_params)
        raise RouteNotFound(f"No model found for path: {path}")



from langchain.embeddings import FakeEmbeddings
from langchain.llms import FakeListLLM

from genoss.entities.chat.chat_completion import ChatCompletion
from genoss.entities.chat.message import Message
from genoss.llm.base_genoss import BaseGenossLLM

FAKE_LLM_NAME = "fake"


class FakeLLM(BaseGenossLLM):
    name: str = FAKE_LLM_NAME
    description: str = "Fake LLM for testing purpose"

    def generate_answer(self, messages: list[Message]) -> ChatCompletion:
        llm = FakeListLLM(responses=["Hello from FakeLLM!"])
        return self._chat_completion_from_langchain_llm(llm=llm, messages=messages)

    def generate_embedding(self, text: str) -> list[float]:
        model = FakeEmbeddings(size=128)
        return model.embed_query(text)



from langchain import PromptTemplate

system_prompt = """\
    You are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe. Answer in the language of the question \
    Here is the question from the user: {question}\
    Answer:\
"""

prompt_template = PromptTemplate(template=system_prompt, input_variables=["question"])



import click
import uvicorn
from uvicorn.config import LOGGING_CONFIG
from uvicorn.main import LEVEL_CHOICES


@click.command(context_settings={"auto_envvar_prefix": "GENOSS"})
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Bind socket to this host.",
    show_default=True,
)
@click.option(
    "--port",
    type=int,
    default=4321,  # Default genoss port
    help="Bind socket to this port. If 0, an available port will be picked.",
    show_default=True,
)
@click.option("--reload", is_flag=True, default=False, help="Enable auto-reload.")
@click.option(
    "--reload-dir",
    "reload_dirs",
    multiple=True,
    help="Set reload directories explicitly, instead of using the current working"
    " directory.",
    type=click.Path(exists=True),
)
@click.option(
    "--reload-include",
    "reload_includes",
    multiple=True,
    help="Set glob patterns to include while watching for files. Includes '*.py' "
    "by default; these defaults can be overridden with `--reload-exclude`. "
    "This option has no effect unless watchfiles is installed.",
)
@click.option(
    "--reload-exclude",
    "reload_excludes",
    multiple=True,
    help="Set glob patterns to exclude while watching for files. Includes "
    "'.*, .py[cod], .sw.*, ~*' by default; these defaults can be overridden "
    "with `--reload-include`. This option has no effect unless watchfiles is "
    "installed.",
)
@click.option(
    "--reload-delay",
    type=float,
    default=0.25,
    show_default=True,
    help="Delay between previous and next check if application needs to be."
    " Defaults to 0.25s.",
)
@click.option(
    "--workers",
    default=None,
    type=int,
    help="Number of worker processes. Defaults to the $WEB_CONCURRENCY environment"
    " variable if available, or 1. Not valid with --reload.",
)
@click.option(
    "--env-file",
    type=click.Path(exists=True),
    default=None,
    help="Environment configuration file.",
    show_default=True,
)
@click.option(
    "--log-config",
    type=click.Path(exists=True),
    default=None,
    help="Logging configuration file. Supported formats: .ini, .json, .yaml.",
    show_default=True,
)
@click.option(
    "--log-level",
    type=LEVEL_CHOICES,
    default=None,
    help="Log level. [default: info]",
    show_default=True,
)
@click.option(
    "--access-log/--no-access-log",
    is_flag=True,
    default=True,
    help="Enable/Disable access log.",
)
@click.option(
    "--use-colors/--no-use-colors",
    is_flag=True,
    default=None,
    help="Enable/Disable colorized logging.",
)
def main(
    host: str,
    port: int,
    reload: bool,
    reload_dirs: list[str],
    reload_includes: list[str],
    reload_excludes: list[str],
    reload_delay: float,
    workers: int,
    env_file: str,
    log_config: str,
    log_level: str,
    access_log: bool,
    use_colors: bool,
) -> None:
    """A simple entrypoint for the default genoss server.

    This is configured in `pyproject.toml:tool.poetry.scripts:genoss-server` .
    You can use it this way
    ```sh
    genoss-server
    genoss-server --help
    genoss-server --log-level debug --reload
    ```

    This is heavily inspired by uvicorn.main / run which have a lot more options.
    To do more, the best solution is to use genoss as it is : a fastapi app.
    You can then configure it as you want and rely on powerful tools provided by
    fastapi, uvicorn or other ASGI tool.
    """
    uvicorn.run(
        app="genoss.default_server:app",
        host=host,
        port=port,
        reload=reload,
        reload_dirs=reload_dirs or None,
        reload_includes=reload_includes or None,
        reload_excludes=reload_excludes or None,
        reload_delay=reload_delay,
        workers=workers,
        env_file=env_file,
        log_config=LOGGING_CONFIG if log_config is None else log_config,
        log_level=log_level,
        access_log=access_log,
        use_colors=use_colors,
    )


if __name__ == "__main__":
    main()



from fastapi import APIRouter

from logger import get_logger

logger = get_logger(__name__)

misc_router = APIRouter()


@misc_router.get("/", tags=["Root"])
async def get_root() -> str:
    return "Genoss API is running!"



from langchain import HuggingFaceHub

from genoss.entities.chat.chat_completion import ChatCompletion
from genoss.entities.chat.message import Message
from genoss.llm.base_genoss import BaseGenossLLM


class BaseHuggingFaceHubLLM(BaseGenossLLM):
    """Class for interacting with Hugging Face Inference APIs."""

    name: str = "HuggingFaceHubLLM"
    description: str = "Hugging Face Hub Inference API"
    api_key: str | None = None
    repo_id: str

    def generate_answer(self, messages: list[Message]) -> ChatCompletion:
        """Generate answer from prompt."""
        llm = HuggingFaceHub(
            repo_id=self.repo_id, huggingfacehub_api_token=self.api_key
        )
        return self._chat_completion_from_langchain_llm(llm=llm, messages=messages)

    def generate_embedding(self, text: str) -> list[float]:
        """Dummy method to satisfy base class requirement."""
        # TODO: why is this necessary? Architecture issue?
        raise NotImplementedError(
            "This method is not used for Hugging Face Inference API."
        )



import logging

from fastapi import FastAPI, HTTPException, Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse

from genoss.api.completions_routes import completions_router
from genoss.api.embeddings_routes import embeddings_router
from genoss.api.misc_routes import misc_router
from logger import get_logger

logger = get_logger(__name__)


app = FastAPI()

app.include_router(completions_router)
app.include_router(embeddings_router)
app.include_router(misc_router)


@app.exception_handler(HTTPException)
async def http_exception_handler(_: Request, exc: HTTPException) -> JSONResponse:
    return JSONResponse(
        status_code=exc.status_code,
        content={"detail": exc.detail},
    )


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(
    request: Request, exc: RequestValidationError
) -> JSONResponse:
    exc_str = f"{exc}".replace("\n", " ").replace("   ", " ")
    logging.error(f"{request}: {exc_str}")
    content = {"status_code": 10422, "message": exc_str, "data": None}
    logger.error(f"Received request: {request}")
    return JSONResponse(
        content=content, status_code=status.HTTP_422_UNPROCESSABLE_ENTITY
    )



from fastapi import APIRouter, Body, HTTPException
from fastapi.params import Depends
from pydantic import BaseModel

from genoss.auth.auth_handler import AuthHandler
from genoss.entities.chat.chat_completion import ChatCompletion
from genoss.entities.chat.message import Message
from genoss.services.chat_completion_registry import chat_completion_registry
from genoss.services.model_routing_helpers import InvalidRouteParameters, RouteNotFound
from logger import get_logger

logger = get_logger(__name__)

completions_router = APIRouter()


class RequestBody(BaseModel):
    model: str
    messages: list[Message]
    temperature: float | None


@completions_router.post("/chat/completions", tags=["Chat Completions"])
async def post_chat_completions(
    # TODO: check if this is the correct way to use Body & Depends
    body: RequestBody = Body(...),  # noqa: B008
    api_key: str = Depends(  # type: ignore[assignment] # noqa: B008
        AuthHandler.check_auth_header, use_cache=False
    ),
) -> ChatCompletion:
    try:
        model = chat_completion_registry.build_model(
            path=body.model, call_params={"api_key": api_key}
        )
    except RouteNotFound as exception:
        logger.exception(f"/chat/completion/ model not found for {body.model}")
        raise HTTPException(status_code=404, detail="Model not found") from exception
    except InvalidRouteParameters as exception:
        logger.exception(f"/chat/completion/ model could not be build for {body.model}")
        raise HTTPException(status_code=404, detail="Model not found") from exception

    logger.info(
        f"Received chat completions request for {model.name} with messages {body.messages[-1].content}"
    )

    # TODO: Add temperature to request body
    return model.generate_answer(body.messages)



import time
import uuid
from typing import Self

from pydantic import BaseModel, Field

from genoss.entities.chat.message import Message


class Choice(BaseModel):
    message: Message
    finish_reason: str = "stop"
    index: int = 0

    @classmethod
    def from_model_answer(cls, answer: str) -> Self:
        return cls(
            message=Message(role="assistant", content=answer),
            finish_reason="stop",
            index=0,
        )


class Usage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int

    @classmethod
    def from_question_and_answer(cls, question: str, answer: str) -> Self:
        return cls(
            prompt_tokens=len(question),
            completion_tokens=len(answer),
            total_tokens=len(question) + len(answer),
        )


class ChatCompletion(BaseModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    object: str = "chat.completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    usage: Usage
    choices: list[Choice]

    @classmethod
    def from_model_question_answer(cls, model: str, question: str, answer: str) -> Self:
        return cls(
            model=model,
            usage=Usage.from_question_and_answer(question, answer),
            choices=[Choice.from_model_answer(answer)],
        )



from langchain.embeddings import GPT4AllEmbeddings
from langchain.llms import GPT4All

from genoss.entities.chat.chat_completion import ChatCompletion
from genoss.entities.chat.message import Message
from genoss.llm.local.base_local import BaseLocalLLM


class Gpt4AllLLM(BaseLocalLLM):
    name: str = "gpt4all"
    description: str = "GPT4ALL"
    model_path: str = "./local_models/ggml-gpt4all-j-v1.3-groovy.bin"

    def generate_answer(self, messages: list[Message]) -> ChatCompletion:
        llm = GPT4All(
            model=self.model_path,  # pyright: ignore reportPrivateUsage=none
        )
        return self._chat_completion_from_langchain_llm(llm=llm, messages=messages)

    def generate_embedding(self, embedding: str | list[str]) -> list[float]:
        gpt4all_embd = GPT4AllEmbeddings()  # pyright: ignore reportPrivateUsage=none
        embedding_to_str = " ".join(embedding)
        return gpt4all_embd.embed_query(embedding_to_str)
